unit MockingMain;

interface

uses
  Winapi.Windows, Winapi.Messages,
  System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls,
  Spring.Mocking;

type
  INotImplemented = interface(IInvokable) ['{6C541309-3D69-454B-B0BB-2E9F4636E5D8}']
    function MyID: integer;
    function MyValues: TArray<integer>;
    procedure ShowMessage;
    function DivideBy(value, quot: integer): real;
    function Fibonacci: integer;
  end;

  TfrmDecorator = class(TForm)
    lbLog: TListBox;
    btnReturns: TButton;
    btnBehavior: TButton;
    btnReturnsArray: TButton;
    btnArguments: TButton;
    btnExecutes: TButton;
    btnAssertion: TButton;
    btnRaises: TButton;
    btnAssertionConditions: TButton;
    btnSpecification: TButton;
    btnSequence: TButton;
    procedure btnBehaviorClick(Sender: TObject);
    procedure btnArgumentsClick(Sender: TObject);
    procedure btnAssertionClick(Sender: TObject);
    procedure btnExecutesClick(Sender: TObject);
    procedure btnRaisesClick(Sender: TObject);
    procedure btnReturnsClick(Sender: TObject);
    procedure btnReturnsArrayClick(Sender: TObject);
    procedure btnSpecificationClick(Sender: TObject);
    procedure btnAssertionConditionsClick(Sender: TObject);
    procedure btnSequenceClick(Sender: TObject);
  private
    function IntArrToStr(const arr: TArray<integer>): string;
    procedure Log(const msg: string); overload;
    procedure Log(const msg: string; const params: array of const); overload;
  public
  end;

var
  frmDecorator: TfrmDecorator;

implementation

uses
  Spring.Collections;

{$R *.dfm}

type
  EMyException = class(Exception)
  end;

procedure TfrmDecorator.btnBehaviorClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create(TMockBehavior.Strict);

  Log('My ID: ' + INotImplemented(mockCalc).MyID.ToString);
end;

procedure TfrmDecorator.btnReturnsClick(Sender: TObject);
var
  mockCalc : Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Returns(17).When.MyID;

  Log('MyID: ' + mockCalc.Instance.MyID.ToString);
end;

procedure TfrmDecorator.btnReturnsArrayClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
  data    : TArray<integer>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  data := [1, 1, 2, 3, 5, 8, 13];
  mockCalc.Setup.Returns([data]).When.MyValues;

  Log('MyValues: ' + IntArrToStr(mockCalc.Instance.MyValues));
end;

procedure TfrmDecorator.btnExecutesClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Executes(
    function (const callInfo: TCallInfo): TValue
    begin
      ShowMessage('Please don''t call this function!');
    end).When.ShowMessage;

  mockCalc.Instance.ShowMessage;
end;

procedure TfrmDecorator.btnRaisesClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Raises<EMyException>('Bad programmer!').When.ShowMessage;

  mockCalc.Instance.ShowMessage;
end;

procedure TfrmDecorator.btnSequenceClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
  sequence: MockSequence;
  i: integer;
begin
  mockCalc := Mock<INotImplemented>.Create(TMockBehavior.Strict);

  mockCalc.Setup(sequence).Returns([1, 2, 3, 5, 8]).When.Fibonacci;

  for i := 1 to 5 do
    Log('Fibonacci: ' + mockCalc.Instance.Fibonacci.ToString);

//  Log('Fibonacci: ' + mockCalc.Instance.Fibonacci.ToString);

  if not sequence.Completed then
    Log('*** Incomplete sequence!');
end;

procedure TfrmDecorator.btnArgumentsClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Executes(
    function (const callInfo: TCallInfo): TValue
    begin
      Result := callInfo.Args[0].AsInteger / callInfo.Args[1].AsInteger;
    end).When.DivideBy(Arg.IsAny<integer>, Arg.IsAny<integer>);

  Log('DivideBy(16,4) = %.0f', [mockCalc.Instance.DivideBy(16,4)]);
end;

procedure TfrmDecorator.btnSpecificationClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Executes(
    function (const callInfo: TCallInfo): TValue
    begin
      Result := callInfo.Args[0].AsInteger / callInfo.Args[1].AsInteger;
    end).When.DivideBy(Arg.IsAny<integer>, Arg.IsAny<integer>);

  mockCalc.Setup.Executes(
    function (const callInfo: TCallInfo): TValue
    begin
      Result := -1; // special kind of arithmetic
    end).When.DivideBy(Arg.IsAny<integer>, Arg.IsEqual<integer>(0));

  mockCalc.Setup.Executes(
    function (const callInfo: TCallInfo): TValue
    begin
      Result := -2; // special kind of arithmetic
    end).When.DivideBy(1,0);

  Log('DivideBy(16,4) = %.0f', [mockCalc.Instance.DivideBy(16,4)]);
  Log('DivideBy(16,0) = %.0f', [mockCalc.Instance.DivideBy(16,0)]);
  Log('DivideBy(16,0) = %.0f', [mockCalc.Instance.DivideBy(1,0)]);
end;

procedure TfrmDecorator.btnAssertionClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Executes.When.MyID;

  mockCalc.Instance.MyID;

  mockCalc.Received.MyID;    // OK
  mockCalc.Received.MyValues; // raise
end; { TfrmDecorator.btnAssertionClick }

procedure TfrmDecorator.btnAssertionConditionsClick(Sender: TObject);
var
  mockCalc: Mock<INotImplemented>;
begin
  mockCalc := Mock<INotImplemented>.Create;

  mockCalc.Setup.Executes.When.DivideBy(Arg.IsAny<integer>, Arg.IsAny<integer>);

  mockCalc.Instance.DivideBy(4,2);
  mockCalc.Instance.DivideBy(5,1);

  mockCalc.Received(Times.AtLeast(2),
    function (const values: TArray<TValue>): boolean
    begin
      Result := values[1].AsInteger > 0;
    end).DivideBy(0,0);
end;

function TfrmDecorator.IntArrToStr(const arr: TArray<integer>): string;
begin
  Result := string.Join(',', TEnumerable.Select<integer,string>(
                               TEnumerable.From<integer>(arr), IntToStr
                             ).ToArray);
end;

procedure TfrmDecorator.Log(const msg: string; const params: array of const);
begin
  Log(Format(msg, params));
end;

procedure TfrmDecorator.Log(const msg: string);
begin
  lbLog.ItemIndex := lbLog.Items.Add(msg);
end;

end.
