unit SpringHttpClient;

interface

uses
  System.TypInfo, System.RTTI, System.SysUtils, System.StrUtils,
  System.Generics.Collections, System.Net.HttpClient,
  System.Net.URLClient,
  Spring, Spring.Collections, Spring.Interception, Spring.Reflection;

type
  AcceptAttribute = class(TCustomAttribute)
  strict private
    FContentType: string;
  public
    /// <summary>Sets Accept: header for the request.</summary>
    /// <summary>AcceptAttribute.Create</summary>
    /// <param name="contentType"> (string) </param>
    /// <param name='contentType'>If major type is ommitted, constructor will default it to
    /// 'application/'. </param>
    constructor Create(const contentType: string);
    property ContentType: string read FContentType write FContentType;
  end; { AcceptAttribute }

  URIPathAttribute = class(TCustomAttribute)
  strict private
    FValue: string;
  public
    constructor Create(const apiPath: string);
    property Value: string read FValue write FValue;
  end;

  THeaderField = TPair<string,string>;
  THeaderFields = IList<THeaderField>;

  TRESTConnectionConfig = record
  private
    FHeader    : THeaderFields;
    FURI       : string;
    FUserAgent : string;
    function GetHeader: THeaderFields;
  public
    property Header: THeaderFields read GetHeader;
    property URI: string read FURI write FURI;
    property UserAgent: string read FUserAgent write FUserAgent;
  end; { TRESTConnectionConfig }

  TRESTConnection<API:IInterface> = record
  private
    FConfig: TRESTConnectionConfig;
  public
    constructor Create(const uri: string);
    function  AcceptXML: TRESTConnection<API>; inline;
    function  Header(const name, value: string): TRESTConnection<API>; inline;
    function  UserAgent(const agentName: string): TRESTConnection<API>; inline;
    function  Invoke: API;
    property Config: TRESTConnectionConfig read FConfig;
  end; { TRESTConnection<I:IInterface> }

  RESTClient<API: IInterface> = record
  public
    class function  Http(const host, path: string): TRESTConnection<API>; overload; static; inline;
    class function  Http(const host: string; port: integer; const path: string): TRESTConnection<API>; overload; static; inline;
    class function  Https(const host, path: string): TRESTConnection<API>; overload; static; inline;
    class function  Https(const host: string; port: integer; const path: string): TRESTConnection<API>; overload; static; inline;
  end; { RESTClient<I> }

  //Defined in interface section because it is used in a parameterized type.
  //Treat it as a private class.
  TRESTInterceptor = class(TInterfacedObject, IInterceptor)
  private
    FConfig  : TRESTConnectionConfig;
    FTypeInfo: PTypeInfo;
  public
    constructor Create(typInfo: PTypeInfo; const config: TRESTConnectionConfig);
    procedure Intercept(const invocation: IInvocation);
  end; { TRESTInterceptor }

  //Defined in interface section because it is used in a parameterized type.
  //Treat it as a private helper.
  RESTHelpers = class abstract
  public
    class function Concat(const host, path: string): string; overload; static;
    class function Concat(const params: TArray<string>): string; overload; static;
  end; { RESTHelpers }

implementation

{ AcceptAttribute }

constructor AcceptAttribute.Create(const contentType: string);
begin
  inherited Create;
  if Pos('/', contentType) = 0 then
    FContentType := 'application/' + contentType
  else
    FContentType := contentType;
end; { AcceptAttribute.Create }

{ URIPathAttribute }

constructor URIPathAttribute.Create(const apiPath: string);
begin
  inherited Create;
  FValue := apiPath;
end; { URIPathAttribute.Create }

{ RESTHelpers }

class function RESTHelpers.Concat(const host, path: string): string;
begin
  Result := host;
  if not (Result.EndsWith('/') or path.StartsWith('/')) then
    Result := Result + '/';
  Result := Result + path;
  if not Result.EndsWith('/') then
    Result := Result + '/';
end; { RESTHelpers.Concat }

class function RESTHelpers.Concat(const params: TArray<string>): string;
var
  i: integer;
begin
  Result := params[0];
  for i := 1 to High(params) do
    Result := Concat(Result, params[i]);
end; { RESTHelpers.Concat }

{ TRESTConnectionConfig }

function TRESTConnectionConfig.GetHeader: THeaderFields;
begin
  if not assigned(FHeader) then
    FHeader := TCollections.CreateList<THeaderField>;
  Result := FHeader;
end; { TRESTConnectionConfig.GetHeader }

{ TRESTConnection<API> }

constructor TRESTConnection<API>.Create(const uri: string);
var
  attr   : URIPathAttribute;
  uriPath: string;
begin
  uriPath := '';
  if assigned(TypeInfo(API))
     and TType.GetType(TypeInfo(API)).TryGetCustomAttribute<URIPathAttribute>(attr)
  then
    uriPath := attr.Value;

  FConfig.URI := RESTHelpers.Concat(uri, uriPath);
end; { TRESTConnection }

function TRESTConnection<API>.AcceptXML: TRESTConnection<API>;
begin
  Result := Header('accept', 'application/xml');
end; { TRESTConnection<API>.AcceptXML }

function TRESTConnection<API>.Header(const name, value: string): TRESTConnection<API>;
begin
  FConfig.Header.Add(THeaderField.Create(name, value));
  Result := Self;
end; { TRESTConnection<API.Header }

function TRESTConnection<API>.Invoke: API;
begin
  Result := TProxyGenerator.CreateInterfaceProxyWithoutTarget<API>(
              TRESTInterceptor.Create(TypeInfo(API), FConfig));
end; { TRESTConnection }

function TRESTConnection<API>.UserAgent(const agentName: string): TRESTConnection<API>;
begin
  FConfig.UserAgent := agentName;
  Result := Self;
end; { TRESTConnection }

{ RESTClient<API> }

class function RESTClient<API>.Http(const host, path: string): TRESTConnection<API>;
begin
  Result := TRESTConnection<API>.Create('http://' + RESTHelpers.Concat(host, path));
end; { RESTClient<I>.Http }

class function RESTClient<API>.Http(const host: string; port: integer; const path:
  string): TRESTConnection<API>;
begin
  Result := TRESTConnection<API>.Create('http://' + RESTHelpers.Concat(Format('%s:%d', [host, port]), path));
end; { RESTClient<I>.Http }

class function RESTClient<API>.Https(const host, path: string): TRESTConnection<API>;
begin
  Result := TRESTConnection<API>.Create('https://' + RESTHelpers.Concat(host, path));
end; { RESTClient<I>.Https }

class function RESTClient<API>.Https(const host: string; port: integer; const path:
  string): TRESTConnection<API>;
begin
  Result := TRESTConnection<API>.Create('htts://' + RESTHelpers.Concat(Format('%s:%d', [host, port]), path));
end; { RESTClient<I>.Https }

{ TRESTInterceptor }

constructor TRESTInterceptor.Create(typInfo: PTypeInfo; const config: TRESTConnectionConfig);
begin
  inherited Create;
  FTypeInfo := typInfo;
  FConfig := config;
end; { TRESTInterceptor.Create }

procedure TRESTInterceptor.Intercept(const invocation: IInvocation);
var
  acceptAttr : AcceptAttribute;
  client     : THTTPClient;
  first      : boolean;
  iParam     : integer;
  paramName  : string;
  params     : TArray<TRttiParameter>;
  paramTag   : string;
  request    : IHTTPRequest;
  response   : IHTTPResponse;
  uri        : string;
  uriPathAttr: URIPathAttribute;
begin
  uri := FConfig.FURI;

  uriPathAttr := invocation.Method.GetCustomAttribute<URIPathAttribute>;
  if assigned(uriPathAttr) then
    uri := uri + uriPathAttr.Value;

  if uri.Contains('{action}') then
    uri := StringReplace(uri, '{action}', invocation.Method.Name.ToLower, [rfReplaceAll])
  else
    uri := RESTHelpers.Concat(uri, invocation.Method.Name.ToLower);

  params := invocation.Method.GetParameters;
  Assert(Length(params) = Length(invocation.Arguments));
  Assert(Low(params) = Low(invocation.Arguments));

  first := true;
  for iParam := Low(params) to High(params) do begin
    paramName := params[iParam].Name.ToLower;
    paramTag := '{' + paramName + '}';
    if uri.ToLower.Contains(paramTag.ToLower) then
      uri := StringReplace(uri, paramTag, invocation.Arguments[iParam].ToString, [rfReplaceAll, rfIgnoreCase])
    else begin
      uri := uri.TrimRight(['/']) + IfThen(first, '?', '&') + paramName + '=' + invocation.Arguments[iParam].ToString;
      first := false;
    end;
  end; //for iParam

  acceptAttr := invocation.Method.GetCustomAttribute<AcceptAttribute>;
  if assigned(acceptAttr) then
    FConfig.Header.Add(THeaderField.Create('accept', acceptAttr.ContentType));

  client := THTTPClient.Create;
  try
    request := client.GetRequest('GET', uri);
    request.UserAgent := FConfig.UserAgent;
    response := client.Execute(request, nil,
      TEnumerable.Select<THeaderField, TNameValuePair>(
        FConfig.Header,
        function (field: THeaderField): TNameValuePair
        begin
          Result := TNameValuePair.Create(field.Key, field.Value);
        end).ToArray);
    invocation.Result := TValue.From<IHTTPResponse>(response);
  finally FreeAndNil(client); end;
end; { TRESTInterceptor.Intercept }

end.
